
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_common.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_env_variable.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_static_process_types.h"
#include "../../src/initng_service.h"

#include "initng_pidfile.h"

/*
 * timeout waiting for a pidfile to be created
 * after the daemon returns happily.
 */
#define PID_TIMEOUT 10

/* Rate limit on missing pidfile warnings */
#define PID_WARN_RATE 2

/* The name of the process, initng should probe */
s_entry PIDOF = { "pid_of", STRING, &TYPE_DAEMON,
    "When daemon exits, initng will look for a process with this name, and set daemon pid no to that pid."
};

/* The filename/path of a pidfile that initng can fetch pid no */
s_entry PIDFILE = { "pid_file", STRINGS, &TYPE_DAEMON,
    "When daemon exits, initng will get pid of daemon from this file."
};

/* Determines when initng looks for a pidfile */
s_entry FORKS = { "forks", STRING, &TYPE_DAEMON,
    "Does the daemon fork?"
};

/*
 * this is an internal option, that is set after the
 * service died once, and the pid is updated.
 */
s_entry INTERNAL_GOT_PID = { NULL, SET, 0, NULL, NULL };
s_entry INTERNAL_PIDFILE_TIMEOUT = { NULL, INT, 0, NULL, NULL };
s_entry INTERNAL_PID_WARN_TIME = { NULL, INT, 0, NULL, NULL };

static int clear_pidfile(active_db_h * s);
static void pid_try(active_db_h * service);
a_state_h WAIT_FOR_PID_FILE = { "WAIT_FOR_PID_FILE", IS_STARTING, &pid_try };

static int check_pidfile_opts(service_cache_h * s)
{
    const char* forks;
    if(initng_service_cache_is(&FORKS,s)) 
    {
        forks = initng_service_cache_get_string(&FORKS,s);
        if(strcmp(forks,"yes") != 0 && strcmp(forks,"no") != 0 &&
	   strcmp(forks,"noret") != 0) 
	{
	    F_("Bad option forks=%s\n", forks);
	    return (FALSE);
	}
	if(strcmp(forks,"yes") == 0 && !initng_service_cache_is(&PIDOF,s)
	   && !initng_service_cache_is(&PIDFILE,s))
	{	   
	    F_("Service %s forks but has no pid_file or pid_of\n", s->name);
	    return (FALSE);
	}
    }
    return (TRUE);
}

static pid_t pid_of(const char *name)
{
    DIR *dir;
    struct dirent *d;
    FILE *fp;

#define BUFF_SIZE 512

    D_("Will check for: %s\n", name);

    dir = opendir("/proc");
    if (!dir)
        return (-1);

    /* Walk through the directory. */
    while ((d = readdir(dir)) != NULL)
    {
        char buf[BUFF_SIZE + 1];
        char *s = NULL;
        int len = 0;
        pid_t pid = -1;

        /* See if this is a process */
        if ((pid = atoi(d->d_name)) <= 0)
            continue;

        /* Open the status file. */
        snprintf(buf, BUFF_SIZE, "/proc/%d/stat", pid);
        D_("To open: %s\n", buf);

        /* Read SID & statname from it. */
        if (!(fp = fopen(buf, "r")))
        {
            W_("Could not open %s.\n", buf);
            continue;
        }

        /* HACK: ignore errors */
        if (fgets(buf, BUFF_SIZE, fp) == NULL)
            buf[0] = '\0';

        s = buf;

        /* skip to the first space */
        while (*s && *s != ' ')
            s++;
        if (*s == '\0')
            continue;

        /* skip the space */
        s++;

        /* skipp the '(' char */
        if (*s != '(')
            continue;
        s++;

        /* count the lenght */
        while (s[len] && s[len] != ')')
            len++;

        if (strncmp(s, name, len) == 0)
        {
            D_("Found %s with pid %d\n", name, pid);
            return (pid);
        }
    }

    D_("Did not find a process with name %s\n", name);
    return (-1);
}


/*
 * When trying to start a service, don't allow it to
 * be set RUNNING if there must exist a pidfile, and
 * pid entry has to be updated.
 */
static int check_if_done(active_db_h * s, a_state_h * status)
{
    assert(s);
    assert(s->name);

    S_;

    /* only check if pidfile exists, if we are trying to start something */
    if (status != &RUNNING)
        return (TRUE);

    if (initng_active_db_is(&INTERNAL_GOT_PID, s))
        return (TRUE);

    /* Make sure the service uses a pidfile */
    if (initng_active_db_is(&PIDFILE, s) || initng_active_db_is(&PIDOF, s))
    {
        if (initng_active_db_is(&FORKS,s) &&
	    strcmp(initng_active_db_get_string(&FORKS, s),"yes") != 0)
	{
	  /* it doesn't fork and return, so don't wait for it to die */
	    initng_common_mark_service(s, &WAIT_FOR_PID_FILE);
	    initng_active_db_set_int(&INTERNAL_PIDFILE_TIMEOUT, s, 0);
	    initng_active_db_set_int(&INTERNAL_PID_WARN_TIME, s, (int) time(0));
	    return (FALSE);
	}
	else 
	{
            /* the service have to die, and check_on_kill have to 
               collect the pid first */
	    return (FALSE);
	}
    }
    return (TRUE);
}

/*
 * Check if a pidfile exists, if it exists, update the 
 * pid in the active_db entry. and return TRUE
 */
static pid_t pid_from_file(const char *name, int warn)
{
    int fd = 0;
    int len = 0;
    char buf[21];

    assert(name);

    /* open pid file */
    fd = open(name, O_RDONLY);

    /* If we cant open pidfile, this is bad */
    if (fd == -1)
    {
        if(warn) 
             W_("Unable to open pidfile: %s, \"%s\", it might not be created yet.\n", name, strerror(errno));
        return (-1);
    }

    /* Read data from buffer */
    len = read(fd, buf, 20);
    close(fd);
    if (len < 1)
    {
        F_("Read 0 chars from %s, It's empty.\n", name);
        return (-1);
    }

    /* remove last newline */
    if (buf[len - 1] == '\n')
        buf[len - 1] = 0;
    else
        buf[len] = 0;

    /* Try to convert pid to int */
    return (atoi(buf));
}

static pid_t get_pidof(active_db_h * s)
{
    pid_t pid;
    const char *pidof;
    char *pidof_fixed = NULL;

    pidof = initng_active_db_get_string(&PIDOF, s);
    if (!pidof)
        return (-1);


    pidof_fixed = fix_variables(pidof, s);
    if (!pidof_fixed)
        return (-1);

    pid = pid_of(pidof_fixed);
    free(pidof_fixed);
    return (pid);
}

/* this will get the pid of PIDFILE entry of service */
static pid_t get_pidfile(active_db_h * s, int warn)
{
    pid_t pid;
    const char *pidfile = NULL;
    char *pidfile_fixed = NULL;

    /* get the pidfile */
    while((pidfile = initng_active_db_get_next_string(&PIDFILE, s, pidfile)))
    {
        /* fix the variables in the string */
        pidfile_fixed = fix_variables(pidfile, s);

	/* check so we got the string */
	if (!pidfile_fixed)
	  return (-1);

	/* get the pid from the file */
	pid = pid_from_file(pidfile_fixed, warn);
	free(pidfile_fixed);

	/* return the pid */
	if(pid > 1)
	  return (pid);
    }
    return (-1);
}

/*
 * When a process dies, probably forked and return,
 * it will get here, this is a good way to check if
 * for a pidfile.
 */
static int check_on_kill(active_db_h * s, process_h * p)
{
    assert(s);
    assert(s->name);
    assert(p);

    S_;

    /* make sure this is all about a daemon */
    if (p->pt != &T_DAEMON && s->type != &TYPE_DAEMON)
        return (FALSE);

    /* make sure this daemon is launching, if it is stopping or up do nothing */
    if (!IS_MARK(s, &STARTING))
        return (FALSE);

    D_("will handle %s status %s\n", s->name, s->current_state->state_name);

    /* if rcode indicate failure, dont check for pidfile */
    if (p->r_code > 0)
    {
        W_("Daemon for service %s returned with an bad return code %i, wont look for a pidfile.\n", s->name, p->r_code);
        return (FALSE);
    }

    if (initng_active_db_is(&PIDOF, s) || initng_active_db_is(&PIDFILE, s))
    {
        /* FIXME: We probably ought to mark it FAIL_STARTING */
        if (initng_active_db_is(&FORKS,s) &&
	  strcmp(initng_active_db_get_string(&FORKS, s),"yes") != 0)
	  return (FALSE);
        initng_common_mark_service(s, &WAIT_FOR_PID_FILE);
	/* set the timeout also */
	initng_active_db_set_int(&INTERNAL_PIDFILE_TIMEOUT, s, (int) time(0) + PID_TIMEOUT);
	initng_active_db_set_int(&INTERNAL_PID_WARN_TIME, s, (int) time(0));
        return (TRUE);
    }

    return (FALSE);
}

static int clear_pidfile(active_db_h * s)
{
    /* if the daemon succeeds or fails, remove the got_pid so it wont exist in next try */
    if (IS_MARK(s, &RUNNING) || IS_MARK(s, &FAIL_STARTING))
        initng_active_db_unset(&INTERNAL_GOT_PID, s);


    /* if there exist a stale pidfile when trying to start, remove it */
    if (IS_MARK(s, &START_DEP_MET))
    {
        const char *pidfile = NULL;
	while((pidfile = initng_active_db_get_next_string(&PIDFILE, s, pidfile))) 
	{
	    if (pidfile)
	    {
	        if (unlink(pidfile) != 0 && errno != ENOENT)
		{
		  F_("Could not remove stale pidfile %s\n", pidfile);
		  return (FALSE);
		}
	    }
        }
    }
    return (TRUE);
}

/*
 * Is run on every interrupt for services WAIT_FOR_PID_FILE */
static void pid_try(active_db_h * s)
{
    pid_t pid = -1;
    process_h *p = NULL;
    int timeout = 0, wt, warn = 0;
    
    
    timeout = initng_active_db_get_int(&INTERNAL_PIDFILE_TIMEOUT, s);
    if(timeout>0 && timeout < time(0))
    {
	F_("Service \"%s\" wait for pidfile timed out!\n", s->name);
        initng_common_mark_service(s, &FAIL_STARTING);
        return;
    }

    wt = initng_active_db_get_int(&INTERNAL_PID_WARN_TIME, s);
    if(wt > time(0) || (wt +  PID_WARN_RATE) < time(0)) 
    {
        warn = 1;
	initng_active_db_set_int(&INTERNAL_PID_WARN_TIME, s, (int) time(0));
    }

    /* get the process to handle */
    p = initng_process_db_get(&T_DAEMON, s);
    if (!p)
    {
        F_("Did not find a daemon process on this service!\n");
        initng_common_mark_service(s, &FAIL_STARTING);
        return;
    }

    /* check if string PIDOF or PIDFILE exits */
    if (initng_active_db_is(&PIDOF, s))
    {
        D_("getting pid by PIDOF!\n");
        /* get pid by process name */
        pid = get_pidof(s);
        D_("result : %d\n", pid);
    }

    if (initng_active_db_is(&PIDFILE, s))
    {
        D_("getting pid by PIDFILE!\n");
        pid = get_pidfile(s, warn);
        D_("result : %d\n", pid);
    }

    if (pid < 2)
    {
        /* make sure this one is run in 1 second again */
        initng_global_set_sleep(1);
        return;
    }

    if (initng_active_db_is(&FORKS,s) &&
	strcmp(initng_active_db_get_string(&FORKS, s),"no") == 0 &&
	pid != p->pid)
    {
        F_("Daemon \"%s\" with fork=no, forked!\n");
        initng_common_mark_service(s, &FAIL_STARTING);
        return;
    }

    if (!initng_active_db_is(&FORKS,s) ||
	strcmp(initng_active_db_get_string(&FORKS, s),"yes") == 0)
        /* set the new pid, to that we got from pidof or pidfile */
        p->pid = pid;

    /* FIXME: if forks=noret, we should send messages to the pid in the
       pidfile (and track changes in it), but use the original pid for
       death detection - needed for mysql */

    /* set INTERNAL_GOT_PID so that this service may be set to done */
    initng_active_db_set(&INTERNAL_GOT_PID, s);
    initng_active_db_unset(&INTERNAL_PIDFILE_TIMEOUT, s);
    
    /* mark this service running */
    initng_common_mark_service(s, &RUNNING);
}

int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_data_types_add(&PIDFILE);
    initng_service_data_types_add(&PIDOF);
    initng_service_data_types_add(&FORKS);
    initng_service_data_types_add(&INTERNAL_GOT_PID);
    initng_service_data_types_add(&INTERNAL_PIDFILE_TIMEOUT);
    initng_active_state_add_test(&RUNNING, 90, &check_if_done);
    initng_plugin_hook_add(HANDLE_KILLED, 10, &check_on_kill);
    initng_plugin_hook_add(ASTATUS_CHANGE, 90, &clear_pidfile);
    initng_plugin_hook_add(ADDITIONAL_PARSE, 80, &check_pidfile_opts);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_service_data_types_del(&PIDFILE);
    initng_service_data_types_del(&PIDOF);
    initng_service_data_types_del(&FORKS);
    initng_service_data_types_del(&INTERNAL_GOT_PID);
    initng_service_data_types_del(&INTERNAL_PIDFILE_TIMEOUT);
    initng_active_state_del_test(&RUNNING, &check_if_done);
    initng_plugin_hook_del(HANDLE_KILLED, &check_on_kill);
    initng_plugin_hook_del(ASTATUS_CHANGE, &clear_pidfile);
    initng_plugin_hook_del(ADDITIONAL_PARSE, &check_pidfile_opts);
}
