
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>
#include <errno.h>

#include "initng_global.h"

#include "initng_static_states.h"

#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_execute.h"
#include "initng_common.h"
#include "initng_load_module.h"
#include "initng_handler.h"
#include "initng_kill_handler.h"
#include "initng_static_data_id.h"
#include "initng_active_state.h"
#include "initng_static_process_types.h"
#include "initng_service.h"

/* function defines */
static void handle_WAITING_FOR_START_DEP(active_db_h * service_to_start);
static void handle_START_DEP_MET(active_db_h * service_to_start);
static void handle_WAITING_FOR_STOP_DEP(active_db_h * service_to_stop);
static void handle_STOP_DEP_MET(active_db_h * service_to_stop);
static void handle_STOPPED(active_db_h * service_to_stop);
static void handle_STOP_MARKED(active_db_h * service);
static void kill_daemon(active_db_h * service_to_stop);


/*
 * a_state_h
 *
 * const char *NAME
 * int is_rought
 * void state_interrupt
 */

a_state_h LOADING = { "LOADING", IS_STARTING, NULL };
a_state_h STARTING = { "STARTING", IS_STARTING, NULL };
a_state_h WAITING_FOR_START_DEP = { "WAITING_FOR_START_DEP", IS_STARTING,
    &handle_WAITING_FOR_START_DEP
};
a_state_h START_DEP_MET = { "START_DEP_MET", IS_STARTING, &handle_START_DEP_MET };
a_state_h FAIL_STARTING = { "FAIL_STARTING", IS_FAILED, NULL };
a_state_h START_DEP_FAILED = { "START_DEP_FAILED", IS_FAILED, NULL };
a_state_h RUNNING = { "RUNNING", IS_UP, NULL };
a_state_h DONE = { "DONE", IS_UP, NULL };
a_state_h STOPPING = { "STOPPING", IS_STOPPING, NULL };
a_state_h WAITING_FOR_STOP_DEP = { "WAITING_FOR_STOP_DEP", IS_STOPPING,
    &handle_WAITING_FOR_STOP_DEP
};
a_state_h STOP_DEP_MET = { "STOP_DEP_MET", IS_STOPPING, &handle_STOP_DEP_MET
};
a_state_h STOP_DEP_FAILED = { "STOP_DEP_FAILED", IS_FAILED, NULL };
a_state_h FAIL_STOPPING = { "FAIL_STOPPING", IS_FAILED, NULL };
a_state_h STOPPED = { "STOPPED", IS_DOWN, &handle_STOPPED };
a_state_h STOP_MARKED = { "STOP_MARKED", IS_STOPPING, handle_STOP_MARKED };
a_state_h FREEING = { "FREEING", IS_DOWN, NULL };
a_state_h WAIT_FOR_CRON = { "WAIT_FOR_CRON", IS_WAITING, NULL };

void initng_static_states_add_default(void)
{

    initng_active_state_add(&LOADING);
    initng_active_state_add(&STARTING);
    initng_active_state_add(&WAITING_FOR_START_DEP);
    initng_active_state_add(&START_DEP_MET);
    initng_active_state_add(&FAIL_STARTING);
    initng_active_state_add(&START_DEP_FAILED);
    initng_active_state_add(&RUNNING);
    initng_active_state_add(&DONE);
    initng_active_state_add(&STOPPING);
    initng_active_state_add(&WAITING_FOR_STOP_DEP);
    initng_active_state_add(&STOP_DEP_MET);
    initng_active_state_add(&STOP_DEP_FAILED);
    initng_active_state_add(&FAIL_STOPPING);
    initng_active_state_add(&STOPPED);
    initng_active_state_add(&STOP_MARKED);
    initng_active_state_add(&FREEING);
    initng_active_state_add(&WAIT_FOR_CRON);

}



    /* D O    S T U F F */

static void handle_STOP_MARKED(active_db_h * service)
{
    initng_common_mark_service(service, &WAITING_FOR_STOP_DEP);
}


/* This is the function that really launches the start-marked services if all deps are up */
static void handle_WAITING_FOR_START_DEP(active_db_h * service_to_start)
{
    active_db_h *current = NULL;

    assert(service_to_start);

    /*
     *    Check so all deps, that needs service_to_start, is up.
     */
    while_active_db(current)
    {
        if (current == service_to_start)
            continue;

        /* if current is not one that it depends on continue */
        if (initng_active_db_dep_on(service_to_start, current) == FALSE)
            continue;

        /* if servuce dep on is starting, wait a bit */
        if (IS_STARTING(current))
            return;

        /* if its not starting or up, mark it START_DEP_FAILED */
        if (!IS_UP(current))
        {
            D_("service %s depending on service %s status %s, aborting.\n",
               service_to_start->name, current->name,
               current->current_state->state_name);
            initng_common_mark_service(service_to_start, &START_DEP_FAILED);
            return;
        }
    }


    if (g.sys_state == STATE_STOPPING)
    {
        initng_common_mark_service(service_to_start, &STOPPED);
        return;
    }

    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        F_("Can't start service, when system status is: %i !\n", g.sys_state);
        return;
    }

    /* set status to START_DEP_MET */
    initng_common_mark_service(service_to_start, &START_DEP_MET);
}


static void handle_START_DEP_MET_daemon(active_db_h * service_to_start)
{
    if (!initng_common_mark_service(service_to_start, &STARTING))
        return;

    switch (initng_execute_launch(service_to_start, &T_DAEMON))
    {
        case FALSE:
        case FAIL:
            F_("Service %s, could not launch a daemon!\n",
               service_to_start->name);
            initng_common_mark_service(service_to_start, &FAIL_STARTING);
            return;                         /* return with status */
        default:
            break;
    }

    /* a daemon is "UP" when its launched */
    initng_common_mark_service(service_to_start, &RUNNING);
}

static void handle_START_DEP_MET_service(active_db_h * service_to_start)
{
    if (!initng_common_mark_service(service_to_start, &STARTING))
        return;

    /* F I N A L L Y   S T A R T */
    switch (initng_execute_launch(service_to_start, &T_START))
    {
        case FALSE:
        case FAIL:
            F_("Service %s, could not launch start, did not find any to launch!\n", service_to_start->name);
            initng_common_mark_service(service_to_start, &FAIL_STARTING);
            return;                         /* return with status */
        default:
            return;
    }
}

static void handle_START_DEP_MET(active_db_h * service_to_start)
{
    assert(service_to_start);

    if (g.sys_state == STATE_STOPPING)
    {
        initng_common_mark_service(service_to_start, &STOPPED);
        return;
    }

    if (service_to_start->type == &TYPE_RUNLEVEL
        || service_to_start->type == &TYPE_VIRTUAL)
    {
        initng_common_mark_service(service_to_start, &DONE);
        return;
    }
    else if (service_to_start->type == &TYPE_SERVICE)
    {
        handle_START_DEP_MET_service(service_to_start);
        return;
    }
    else if (service_to_start->type == &TYPE_DAEMON)
    {
        handle_START_DEP_MET_daemon(service_to_start);
        return;
    }

    F_("DEP MET BUT CANT HANDLE!\n");
    initng_common_mark_service(service_to_start, &FAIL_STARTING);
}


/*
 * This function is called on interrupt for all services with
 * active status WAITING_FOR_STOP_DEP.
 */
static void handle_WAITING_FOR_STOP_DEP(active_db_h * service_to_stop)
{
    active_db_h *current = NULL;

    assert(service_to_stop);
    D_("handle_WAITING_FOR_STOP_DEP (%s)!\n", service_to_stop->name);
    /*
     *    Check so all deps, that needs service_to_stop, is down.
     *    if there are services depending on this one still running, return false and still try
     */
    while_active_db(current)
    {
        if (current == service_to_stop)
            continue;

        /* Does service_to_stop depends on current ?? */
        if (initng_active_db_dep_on(current, service_to_stop) == FALSE)
            continue;

        /* if its done, this is pefrect */
        if (IS_DOWN(current))
            continue;

        /* If the dep is failed, continue */
        if (IS_FAILED(current))
            continue;

        /* else RETURN */
        D_("still waiting for %s state %s\n", current->name,
           current->current_state->state_name);
        return;
    }

    /* this wont execute anything ... */
    if (service_to_stop->type == &TYPE_RUNLEVEL ||
        service_to_stop->type == &TYPE_VIRTUAL)
    {
        initng_common_mark_service(service_to_stop, &STOPPED);
        return;
    }

    /* ok, stopping deps are met */
    D_("Try set STOP_DEP_MET!\n");
    initng_common_mark_service(service_to_stop, &STOP_DEP_MET);
}





/*
 *   Handles:
 *     STOP_DEP_MET   ----->>   STOPPING
 */

static void handle_STOP_DEP_MET_service(active_db_h * service_to_stop)
{
    /* mark this service as STOPPING */
    initng_common_mark_service(service_to_stop, &STOPPING);

    /* launch stop service */
    switch (initng_execute_launch(service_to_stop, &T_STOP))
    {
        case FAIL:
            F_("  --  (%s): fail launch stop!\n", service_to_stop->name);
            initng_common_mark_service(service_to_stop, &FAIL_STOPPING);
            return;
        case FALSE:
            /* there exists no stop process */
            initng_common_mark_service(service_to_stop, &STOPPED);
            return;
        default:
            /* the alarm makes sure this will be killed if it takes to long. */
            if (initng_active_db_is(&STOP_TIMEOUT, service_to_stop))
            {
                initng_global_set_sleep(initng_active_db_get_int
                                        (&STOP_TIMEOUT, service_to_stop) + 1);
            }
            else
            {
                initng_global_set_sleep(SECONDS_BEFORE_KILL + 1);
            }
            return;
    }
}

static void handle_STOP_DEP_MET_daemon(active_db_h * service_to_stop)
{
    process_h *process, *safe = NULL;

    /* mark this service as STOPPING */
    initng_common_mark_service(service_to_stop, &STOPPING);

    /* find the daemon, and check so it still exits */
    while_processes_safe(process, safe, service_to_stop)
    {
        if (process->pt != &T_DAEMON)
            continue;

        if (process->pid <= 0)
        {
            initng_common_mark_service(service_to_stop, &STOPPED);
            return;
        }

        if (kill(process->pid, 0) && errno == ESRCH)
        {
            initng_common_mark_service(service_to_stop, &STOPPED);
            return;
        }
    }

    /* launch stop service */
    switch (initng_execute_launch(service_to_stop, &T_KILL))
    {
        case FAIL:
            F_("  --  (%s): fail launch stop!\n", service_to_stop->name);
            initng_common_mark_service(service_to_stop, &FAIL_STOPPING);
            return;
        case FALSE:
            /* if there is no plugin that wanna kill this daemon,
             * we do it ourself.
             */
            kill_daemon(service_to_stop);
            break;
        default:
            break;
    }

    if (initng_active_db_is(&STOP_TIMEOUT, service_to_stop))
    {
        initng_global_set_sleep(initng_active_db_get_int
                                (&STOP_TIMEOUT, service_to_stop) + 1);
    }
    else
    {
        initng_global_set_sleep(SECONDS_BEFORE_KILL + 1);
    }

}

static void handle_STOP_DEP_MET(active_db_h * service_to_stop)
{
    if (service_to_stop->type == &TYPE_SERVICE)
    {
        handle_STOP_DEP_MET_service(service_to_stop);
        return;
    }

    if (service_to_stop->type == &TYPE_DAEMON)
    {
        handle_STOP_DEP_MET_daemon(service_to_stop);
        return;
    }

    if (service_to_stop->type == &TYPE_VIRTUAL)
    {
        initng_common_mark_service(service_to_stop, &STOPPED);
        return;
    }

    initng_common_mark_service(service_to_stop, &FAIL_STOPPING);
    F_("DONT KNOW HOW TO STOP %s!\n", service_to_stop->name);
}

static void kill_daemon(active_db_h * service_to_stop)
{
    process_h *process = NULL;

    assert(service_to_stop);

    /* look for a T_DAEMON process */
    while_processes(process, service_to_stop)
    {
        if (process->pt == &T_DAEMON)
            break;
    }

    /* make sure we got a process */
    if (!process)
    {
        F_("Service doesn't have any processes, don't know how to kill then.\n");
        return;
    }

    /* check so we got a T_DAEMON */
    if (process->pt != &T_DAEMON)
    {
        F_("This daemon does not have a daemon_process!\n");
        /*list_del(&process->list);
           process_db_free(process); */
        return;
    }

    /* check so that pid is good */
    if (process->pid <= 0)
    {
        F_("Bad PID %d in database!\n", process->pid);
        list_del(&process->list);
        initng_process_db_free(process);
        return;
    }

    /* check so there exits an process with this pid */
    if (kill(process->pid, 0) && errno == ESRCH)
    {
        F_("Trying to kill a service (%s) with a pid (%d), but there exists no process with this pid!\n", service_to_stop->name, process->pid);
        list_del(&process->list);
        initng_process_db_free(process);
        return;
    }

    /* if system is not stopping, generate a warning */
    if (g.sys_state != STATE_STOPPING)
    {
        W_(" Sending the process %i of %s, the SIGTERM signal!\n",
           process->pid, service_to_stop->name);
    }

    /* Uhm, this doesnt work : kill(-service_to_stop->start_process->pid, SIGTERM); */
    kill(process->pid, SIGTERM);

    if (initng_active_db_is(&STOP_TIMEOUT, service_to_stop))
    {
        initng_global_set_sleep(initng_active_db_get_int
                                (&STOP_TIMEOUT, service_to_stop) + 1);
        D_("Clock is set, killed process %i (%s), have %i seconds to quit before SIGKILL\n", process->pid, service_to_stop->name, initng_active_db_get_int(&STOP_TIMEOUT, service_to_stop));
    }
    else
    {
        initng_global_set_sleep(SECONDS_BEFORE_KILL + 1);
        D_("Clock is set, killed process %i (%s), have %i seconds to quit before SIGKILL\n", process->pid, service_to_stop->name, SECONDS_BEFORE_KILL);
    }
}

static void handle_STOPPED(active_db_h * service_stopped)
{
    active_db_h *current, *safe = NULL;

    /*
     * Make sure there is no services that needs this
     * that still think its running.
     */
    while_active_db_safe(current, safe)
    {
        /* no idea to stop myself */
        if (current == service_stopped)
            continue;

        /* DONT stop runlevels OR virtuals */
        if (current->type == &TYPE_VIRTUAL)
            continue;

        /* check that current needs service_stopped */
        if (initng_active_db_dep_on(current, service_stopped) == FALSE)
            continue;

        /* dont stop a stopped service */
        if (IS_DOWN(current))
            continue;

        /* if stop this */
        D_("%s have to stop %s.\n", service_stopped->name, current->name);
        initng_handler_stop_service(current);
    }

    /* check if this service is restarting */
    if (initng_active_db_is(&RESTARTING, service_stopped))
    {
        initng_active_db_remove(&RESTARTING, service_stopped);
        initng_handler_start_service(service_stopped);
        D_("Service is restarting now!\n");
        return;
    }


    /* free service, and forget */
    initng_active_db_del(service_stopped);
    initng_active_db_free(service_stopped);
    D_("Service removed.\n");
}
