#include <CUnit/CUnit.h>
#include <CUnit/Basic.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include "test_initng.h"

#include "../src/initng_open_read_close.h"

static char *fill_file(const char *filename, int size)
{
    FILE *fp;
    char *buf;
    int i;
    int n;

    /* open file */
    fp = fopen(filename, "w");
    if (fp == NULL)
    {
        fprintf(stderr, "%s:%d Error opening \"%s\" for writing\n",
                __FILE__, __LINE__, filename);
        exit(1);
    }

    /* build our output buffer, filled with random characters */
    buf = (char *) malloc(size + 1);
    if (buf == NULL)
    {
        fprintf(stderr, "%s:%d Out of memory\n", __FILE__, __LINE__);
        exit(1);
    }
    srand48(time(NULL));
    i = 0;
    while (i < size)
    {
        n = lrand48() & 0xFF;
        if (isprint(n))
        {
            buf[i++] = n;
        }
    }
    buf[size] = '\0';

    /* output until we're done */
    if (fwrite(buf, 1, size, fp) != size)
    {
        fprintf(stderr, "%s:%d Error writing file\n", __FILE__, __LINE__);
        exit(1);
    }

    /* close */
    fclose(fp);

    /* return our random buffer */
    return buf;
}

/* 
 * Tests the open_read_close(), and as many failure modes as possible.
 * Verifies that the data read is correct by creating files with 
 * random contents.
 *
 * Missing tests for failure of:
 *
 * - fstat() return error
 * - read() return error
 * - read() not get enough bytes
 * - close() return error
 */
static void test_open_read_close(void)
{
    char *buf;
    char *cmp_buf;

    NEW_TEST();

    TEST(" verify that the read fails if we give it a bad filename ");
    buf = "";
    CU_ASSERT(open_read_close("no-such-filename", &buf) == FALSE);
    CU_ASSERT(buf == NULL);

    /* XXX: how to get fstat() to fail? */

    TEST(" try with an empty file ");
    cmp_buf = fill_file("test_initng.tmp", 0);
    CU_ASSERT(open_read_close("test_initng.tmp", &buf) == TRUE);
    CU_ASSERT(buf != NULL);
    CU_ASSERT(strlen(buf) == 0);
    free(cmp_buf);

    TEST(" try with a non-empty file ");
    cmp_buf = fill_file("test_initng.tmp", 100);
    CU_ASSERT(open_read_close("test_initng.tmp", &buf) == TRUE);
    CU_ASSERT(buf != NULL);
    CU_ASSERT(strlen(buf) == 100);
    CU_ASSERT(strcmp(buf, cmp_buf) == 0);
    free(cmp_buf);

    TEST(" try with a big file ");
    cmp_buf = fill_file("test_initng.tmp", 100000);
    CU_ASSERT(open_read_close("test_initng.tmp", &buf) == TRUE);
    CU_ASSERT(buf != NULL);
    CU_ASSERT(strlen(buf) == 100000);
    CU_ASSERT(strcmp(buf, cmp_buf) == 0);
    free(cmp_buf);

    INFO(" cleanup ");
    if (remove("test_initng.tmp") != 0)
    {
        fprintf(stderr, "%s:%d Error %d removing file; %s\n",
                __FILE__, __LINE__, errno, strerror(errno));
        exit(1);
    }
}


void test_open_read_close_setup(void)
{
    CU_pSuite suite;

    suite = CU_add_suite("open_read_close", empty_init, empty_cleanup);
    CU_add_test(suite, "open_read_close", test_open_read_close);
}
